@@ -8,6 +8,7 @@ import re
 from datetime import datetime, timedelta
 from typing import Optional, Dict, Tuple, Iterable, List
 from concurrent.futures import ThreadPoolExecutor, as_completed
+
 import s3fs
 import numpy as np
 import xarray as xr
@@ -29,10 +30,6 @@ DEFAULT_MAX_TIME_DIFF = timedelta(minutes=30)
 DEFAULT_OUT_ROOT = "images/goes16_l1b"
 DEFAULT_ASH_RGB_ROOT = "images/goes16_ash_rgb"
 
-""" 
-FULL_DISK_LAT_BOUNDS = (-60.0, 60.0)
-FULL_DISK_LON_BOUNDS = (-135.0, -15.0) """
-
 DEFAULT_ASH_LAT_BOUNDS = (-60.0, 40.0)
 DEFAULT_ASH_LON_BOUNDS = (-135.0, -15.0)
 DEFAULT_ASH_RES_DEG = 0.02  # ~2 km
@@ -40,6 +37,9 @@ DEFAULT_ASH_RES_DEG = 0.02  # ~2 km
 DEBUG_GOES16 = True
 SCAN_RE = re.compile(r"_s(\d{4})(\d{3})(\d{2})(\d{2})(\d{2})")
 
+# Global cache for S3 filesystem
+_S3_FS = None
+
 
 def _log(msg: str):
     if DEBUG_GOES16:
@@ -80,10 +80,18 @@ def extract_scan_time_from_name(name: str) -> Optional[datetime]:
 # -----------------------------------------------------------------------------
 
 def _build_s3_filesystem():
-    return s3fs.S3FileSystem(
-        anon=True,
-        config_kwargs={"max_pool_connections": 64},
-    )
+    """
+    Build or reuse a global anonymous S3 filesystem with a larger connection pool.
+    """
+    global _S3_FS
+    if _S3_FS is None:
+        _S3_FS = s3fs.S3FileSystem(
+            anon=True,
+            config_kwargs={"max_pool_connections": 128},
+            default_fill_cache=False,
+        )
+        _log("Initialized global S3 filesystem")
+    return _S3_FS
 
 
 def _candidate_hour_prefixes(target, max_time_diff, bucket, product):
@@ -109,6 +117,36 @@ def _candidate_hour_prefixes(target, max_time_diff, bucket, product):
     return prefixes
 
 
+def _glob_one(fs, prefix: str, ch: str, product: str) -> List[str]:
+    """
+    Find candidate keys for a given prefix/channel.
+
+    Try the canonical pattern first; only fall back to a loose pattern if needed.
+    """
+    out: List[str] = []
+
+    if not prefix.endswith("/"):
+        prefix = prefix + "/"
+
+    # Canonical GOES-16 naming pattern
+    pat_main = f"{prefix}OR_{product}-M6C{ch}_*.nc"
+    try:
+        out = fs.glob(pat_main)
+        if out:
+            return out
+    except FileNotFoundError:
+        pass
+
+    # Fallback, slightly looser pattern
+    pat_fallback = f"{prefix}*{product}-M6C{ch}_*.nc"
+    try:
+        out = fs.glob(pat_fallback)
+    except FileNotFoundError:
+        out = []
+
+    return out
+
+
 # -----------------------------------------------------------------------------
 # Find matching L1b files
 # -----------------------------------------------------------------------------
@@ -125,24 +163,13 @@ def find_ash_keys_for_datetime(
     fs = _build_s3_filesystem()
     prefixes = _candidate_hour_prefixes(target, max_time_diff, bucket, product)
 
-    ch_to_candidates = {ch: [] for ch in channels}
+    ch_to_candidates: Dict[str, List[str]] = {ch: [] for ch in channels}
 
-    def _glob_one(prefix, ch):
-        patterns = [
-            f"{prefix}OR_{product}-M6C{ch}_*.nc",
-            f"{prefix}*{product}-M6C{ch}_*.nc",
-        ]
-        out = []
-        for pat in patterns:
-            try:
-                out.extend(fs.glob(pat))
-            except FileNotFoundError:
-                pass
-        return out
-
-    with ThreadPoolExecutor(max_workers=32) as ex:
+    # Parallel globbing across prefixes *and* channels
+    max_workers = min(32, max(1, len(prefixes) * len(channels)))
+    with ThreadPoolExecutor(max_workers=max_workers) as ex:
         futures = {
-            ex.submit(_glob_one, p if p.endswith("/") else p + "/", ch): ch
+            ex.submit(_glob_one, fs, p, ch, product): ch
             for p in prefixes for ch in channels
         }
 
@@ -150,10 +177,10 @@ def find_ash_keys_for_datetime(
             ch = futures[fut]
             try:
                 ch_to_candidates[ch].extend(fut.result())
-            except Exception:
-                pass
+            except Exception as e:
+                _log(f"Glob failed for band {ch}: {e}")
 
-    keys = {}
+    keys: Dict[str, str] = {}
     for ch, candidates in ch_to_candidates.items():
         if not candidates:
             raise FileNotFoundError(f"No files found for band {ch}")
@@ -176,7 +203,7 @@ def find_ash_keys_for_datetime(
             raise FileNotFoundError(f"No valid file for band {ch}")
 
         keys[ch] = best_file
-        _log(f"Band {ch}: using file {os.path.basename(best_file)}")
+        _log(f"Band {ch}: using file {os.path.basename(best_file)} (Δt={best_diff})")
 
     return keys, target
 
@@ -185,6 +212,19 @@ def find_ash_keys_for_datetime(
 # Download L1b files (always used)
 # -----------------------------------------------------------------------------
 
+def _download_one_band(fs, ch: str, key: str, local_path: str):
+    """
+    Helper for parallel band download.
+    """
+    if os.path.exists(local_path):
+        _log(f"Using cached file for band {ch}: {local_path}")
+        return ch, local_path
+
+    _log(f"Downloading band {ch} → {local_path}")
+    fs.get(key, local_path)
+    return ch, local_path
+
+
 def download_ash_bands_for_datetime(
     year, month, day, hhmm,
     *,
@@ -207,18 +247,19 @@ def download_ash_bands_for_datetime(
     out_dir = os.path.join(out_root, f"{year:04d}", f"{month:02d}", f"{day:02d}", hhmm)
     os.makedirs(out_dir, exist_ok=True)
 
-    local_paths = {}
-    for ch, key in keys.items():
-        fname = os.path.basename(key)
-        local_path = os.path.join(out_dir, fname)
+    local_paths: Dict[str, str] = {}
 
-        if os.path.exists(local_path):
-            _log(f"Using cached file for band {ch}: {local_path}")
-        else:
-            _log(f"Downloading band {ch} → {local_path}")
-            fs.get(key, local_path)
+    # Parallel downloads for all bands
+    with ThreadPoolExecutor(max_workers=len(keys)) as ex:
+        futures = []
+        for ch, key in keys.items():
+            fname = os.path.basename(key)
+            local_path = os.path.join(out_dir, fname)
+            futures.append(ex.submit(_download_one_band, fs, ch, key, local_path))
 
-        local_paths[ch] = local_path
+        for fut in as_completed(futures):
+            ch, local_path = fut.result()
+            local_paths[ch] = local_path
 
     return local_paths, target_dt
 
@@ -415,8 +456,9 @@ def _downsample_to_match(high_res: np.ndarray, low_res: np.ndarray) -> np.ndarra
     return high_res
 
 
-
-from typing import Tuple
+# -----------------------------------------------------------------------------
+# NEW: Geolocations of each pixel on the Ash RGB grid
+# -----------------------------------------------------------------------------
 
 def get_ash_rgb_pixel_geolocations(
     lat_bounds: Tuple[float, float] = DEFAULT_ASH_LAT_BOUNDS,
@@ -495,6 +537,7 @@ def _to_true_color(red: np.ndarray,
     rgb = np.dstack([red, green, blue]).astype(np.float32)
     return np.clip(rgb, 0.0, 1.0)
 
+
 def build_truecolor_rgb_from_paths(
     paths,
     *,
@@ -502,7 +545,7 @@ def build_truecolor_rgb_from_paths(
     lon_bounds=DEFAULT_ASH_LON_BOUNDS,
     res_deg=DEFAULT_ASH_RES_DEG,
     gamma: float = 2.2,
-    ):
+):
     """
     Build a true-color RGB image from GOES-16 ABI L1b visible bands.
 
@@ -557,6 +600,7 @@ def build_truecolor_rgb_from_paths(
     rgb = np.flip(rgb, axis=1)
     return rgb
 
+
 # -----------------------------------------------------------------------------
 # High-level function (ALWAYS downloads)
 # -----------------------------------------------------------------------------
